"""
Trace through common causes of failure in the new archive system.
Prints diagnostics that can be used to help address problems.

Author: Daniel K Lyons <dlyons@nrao.edu>
Since: 2017-09-29

Some other stuff to consider checking:

BDF/SDM completion unnoticed

Multicast messaging trouble
REALfast does not send multicast messages!
BDF/SDM completion signal not received by Amygdala
Multicast disconnected

Amygdala not sending workflow start message

Workflow not responding to workflow start message

Workflow is out of worker threads
Ingest job not sent to cluster
Workflow bug

Cluster not launching ingestion job

Cluster is under too much load, no free processing resources
Cluster offline

Ingestion job hung
Ingestion job failed

Required files not present
NGAS not responding
NGAS full
NGAS server offline or unreachable
"""
import json

from collections import defaultdict
import os
import pendulum
import subprocess
import time
import pika
import requests
from pycapo import CapoConfig
import psycopg2 as pg
import logging

logger = logging.getLogger(__name__)

SECONDS_IN_ONE_DAY = 86400

class FaultChecker:
    def __init__(self, profile):
        self.capo = CapoConfig(profile=profile)

    def __repr__(self):
        repr_str_list = ['{c}(']
        temp = ['{0}={{s.{0}}}'.format(key) for key in vars(self)]
        self_list = ', '.join(temp)
        repr_str_list.append(self_list)
        repr_str_list.append(')')

        return ''.join(repr_str_list).format(c=self.__class__.__name__, s=self)


class ServicesChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        self.tomcat = self.capo.settings('tomcat.deployment')
        self.amqp = self.capo.settings('edu.nrao.archive.configuration.AmqpServer')

        self.running_servlets = self.update_running_servlets()

    def update_running_servlets(self):
        r = requests.get(self.tomcat.uri + '/list',
                         auth=(self.tomcat.username,
                               self.tomcat.password))
        return set(x.split(':')[3]
                   for x
                   in r.text.split('\n')[1:-1]
                   if x.split(':')[1] == 'running')

    def check_servlet(self, servlet):
        r""" Check a given servlet to see if it's running.

        :param servlet: The name of the servlet to check on tomcat.
        :return: True if the servlet is running.
        """
        return servlet in self.running_servlets

    def check_amqp_online(self):
        r""" Check to see if we can connect to AMQP and send a message to ourselves.

        :return: True if a loopback message was send and received on amqp.
        """
        connection = pika.BlockingConnection(
                pika.ConnectionParameters(self.amqp.hostname,
                                          credentials=pika.PlainCredentials(
                                                  self.amqp.username,
                                                  self.amqp.password)))
        channel = connection.channel()
        channel.exchange_declare('faulttrace', exchange_type='fanout',
                                 durable=False, auto_delete=True)
        queue = channel.queue_declare('', durable=False, auto_delete=True, exclusive=True).method.queue
        channel.queue_bind(queue, 'faulttrace')
        channel.basic_publish('faulttrace', '', b'hello, world')
        _, _, message = channel.basic_get(queue)
        channel.queue_delete(queue)
        channel.exchange_delete('faulttrace')
        channel.close()
        connection.close()
        return message == b'hello, world'


class QueuesChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        self.amqp = self.capo.settings('edu.nrao.archive.configuration.AmqpServer')
        self.workflow = self.capo.settings('edu.nrao.archive.workflow.config.WorkflowManagerSettings')

    def get_queue_msg_count(self, queue_name):
        r""" Get the number of messages in the given queue

        :param queue_name: The amqp queue to check.
        :return: The number of messages
        """
        r = requests.get('http://{}:15672/api/queues/%2F/{}'.format(self.amqp.hostname, queue_name),
                         auth=(self.amqp.username, self.amqp.password))
        result = json.loads(r.text)
        return result['messages']


class LastProcessedChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        # let's connect to the database out here
        self.databaseCreds = self.capo.settings('metadataDatabase')
        self.host, self.dbname = self.databaseCreds.jdbcUrl.split(':')[2][2:].split('/')
        self.conn = pg.connect(dbname=self.dbname, host=self.host,
                               user=self.databaseCreds.jdbcUsername,
                               password=self.databaseCreds.jdbcPassword)

        # TODO: Will we need to manage a cursor if we put it in the object?
        self.cursor = self.conn.cursor()

    def get_last_ingested(self, product_type):
        r""" Check when we last ingested a <product_type> file

        :param product_type: The type of file we want to check the date on.
        :return: A pendulum instance of the ingest time
        """
        sql = "select max(ingestion_time) from files where format = %s"
        self.cursor.execute(sql, (product_type,))
        ingest_time, = self.cursor.fetchone()
        return pendulum.instance(ingest_time)

        # Check out the newest execution block we have

    def get_last_eb(self):
        r""" Check when the last eb ran.

        :return: A pendulum instance of the time the last eb ran.
        """
        sql = "select mjd_to_timestamp(max(endtime)) from execution_blocks"
        self.cursor.execute(sql)
        eb_time, = self.cursor.fetchone()
        return pendulum.instance(eb_time)

    def get_last_alma_reingestion(self):
        r""" Check when the last ALMA reingestion happened.

        :return: A tuple of (remaining, last_ingested) where remaining
        is the number of ALMA SDMs remaining to be processed
        """
        sql = """select coalesce(max(last_updated), '2018-01-01') 
        from alma_reingestion_queue where state IN ('COMPLETE', 'FAILED')"""
        self.cursor.execute(sql)
        last_updated, = self.cursor.fetchone()

        sql = "select count(*) from alma_reingestion_queue where state IN ('QUEUED', 'WAITING')"
        self.cursor.execute(sql)
        remaining, = self.cursor.fetchone()

        return remaining, pendulum.instance(last_updated)


class DirChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        self.sdmpath = self.capo['archive-ingestion.SDMPath']
        self.mchammer = '/home/mchammer/evla/mcaf/workspace/'
        # let's connect to the database out here
        self.databaseCreds = self.capo.settings('metadataDatabase')
        self.host, self.dbname = self.databaseCreds.jdbcUrl.split(':')[2][2:].split('/')
        self.conn = pg.connect(dbname=self.dbname, host=self.host,
                               user=self.databaseCreds.jdbcUsername,
                               password=self.databaseCreds.jdbcPassword)

        # TODO: Will we need to manage a cursor if we put it in the object?
        self.cursor = self.conn.cursor()

    def get_missing_sdms(self):
        r""" Return a list of missing sdms.

        :return: A list of missing sdms, or an empty list.
        """
        now = time.time()

        sdms = [f.name for f in os.scandir(self.sdmpath)
                if (now - os.stat(os.path.join(self.sdmpath, f.name)).st_mtime < 7 * SECONDS_IN_ONE_DAY) and (f.is_dir())]
        self.cursor.execute('SELECT ngas_fileset_id FROM execution_blocks WHERE ngas_fileset_id IN %s', [tuple(sdms)])
        missing_sdms = set(sdms) - set(r[0] for r in self.cursor.fetchall())

        return [f for f in missing_sdms]

    def get_mchammer_workspace(self):
        r""" A test stub for returning the contents of a directory

        :return: Return the contents of self.ls_test_dir
        """
        # dir_contents = [f for f in os.listdir(self.mchammer)]
        # return dir_contents

        now = time.time()
        sdms = [f for f in os.listdir(self.mchammer)
                if now - os.stat(os.path.join(self.mchammer, f)).st_mtime < 7 * SECONDS_IN_ONE_DAY]
        self.cursor.execute('SELECT ngas_fileset_id FROM execution_blocks WHERE ngas_fileset_id IN %s', [tuple(sdms)])
        missing_sdms = set(sdms) - set(r[0] for r in self.cursor.fetchall())
        return [f for f in missing_sdms]


class ClusterChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        # Username and password for accessing lustre & cluster information (use the deployment info)
        self.wf_deploy = self.capo.settings('workflow.deployment')
        self.uname = self.wf_deploy.username
        self.rhost = self.wf_deploy.hostname

        self.usage_connector = self.uname+'@'+self.rhost

    def check_cluster_usage(self):
        """A utility wrapping the qstat command which will take the list
         of what is running on the cluster in the 'batch' queue and
         provide a relevant subset of the data.

         Potential extensions: Some larger scale statistics (after we agree
         on a set of keys
            % of jobs that are vlapipe vs others
            % of queued jobs (>9% means potential over subscription)

            multiple queues (test, vlass)
            highlight distinctions between VLASS, Old Calibration, and
              AAT/PPI jobs by name.

         :returns A dictionary of jobID:dictionary_of_values
                    where dictionary_of_values has the keys:
                        name, status, node, timeUsed

                or None if an error is encountered.
         """
        # First, grab the data we need and do basic transformations:
        try:
            capture = subprocess.check_output(['ssh',
                                               self.usage_connector,
                                               '/opt/services/torque/bin/qstat',
                                               '-1n',
                                               'batch'])
        except subprocess.CalledProcessError:
            # print("Cannot execute qstat command: " + ex.stderr)
            # something went off the rails.
            return None
        # all is well
        text = capture.decode()
        output_lines = text.splitlines()

        # Loop over the output lines.  Accumulate for overall statistics and
        # populate the individual fields with desirable data.
        cluster_jobs = defaultdict()

        # our usage %
        vlapipe_job_count = 0
        others_job_count = 0

        # % queued, an indication of over subscription
        jobs_queued_count = 0
        jobs_total_count = 0

        for data_line in output_lines[11:]:  # First 6 lines are the NRAO header, then 5 lines of qstat header
            current_job = defaultdict()
            fields = data_line.split()    # default splits on whitespace
            #
            #  0 - ID
            #  1 - uname                (statistics, but not forwarded)
            #  2 - queue                (discard)
            #  3 - jobname
            #  4 - SessionID            (discard)
            #  5 - Nodes requested      (discard)
            #  6 - Tasks                (discard)
            #  7 - Memory requested     (discard)
            #  8 - Time requested       (discard, or provide a %)
            #  9 - State: (R,Q,W,etc)
            # 10 - Time Elapsed
            # 11 - Node(s) used
            #
            current_job['name'] = fields[3]
            current_job['status'] = fields[9]
            if 'Q' == fields[9]:
                jobs_queued_count += 1
            current_job['nodes'] = fields[11]
            current_job['timeUsed'] = fields[10]

            # add this job to the overall result:
            job_id_number = fields[0].split('.')[0]
            cluster_jobs[job_id_number] = current_job

            if 'vlapipe' == fields[1]:
                vlapipe_job_count += 1
            else:
                others_job_count += 1
            # endif

            jobs_total_count += 1
        # endfor

        #
        # A few overall metrics in case we're interested.
        #

        # percent_current_usage = vlapipe_job_count/(others_job_count+vlapipe_job_count)*100.0
        # clusterJobs['usagePercent'] = percent_current_usage

        # percent_queued_jobs = jobs_queued_count/jobs_total_count*100.0
        # clusterJobs['queuedPercent'] = percent_queued_jobs

        return cluster_jobs


class LustreChecker(FaultChecker):
    def __init__(self, profile):
        super().__init__(profile)
        # Username and password for accessing lustre & cluster information (use the deployment info)
        self.wf_deploy = self.capo.settings('workflow.deployment')
        self.uname = self.wf_deploy.username
        self.rhost = self.wf_deploy.hostname

        self.usage_connector = self.uname+'@'+self.rhost

    def lustre_usage(self):
        """ A wrapper of the lustre df and quota utilities to
            provide an overall view of processing space usage.
            This will grab the total size of lustre, the amount
            which is used, and the amount used by the vlapipe
            group, and provide them via a dictionary.

            If everything goes to plan, it will return a dictionary of
            disk space values, all in terms of 1k blocks. With the keys:

                total:
                used:
                vlapipe:

            On error, returns None
        """
        try:
            capture = subprocess.check_output(['ssh', self.usage_connector, 'df', '/lustre/aoc'])
        except subprocess.CalledProcessError:
            # something went off the rails.
            return None
        # all is well
        text = capture.decode()
        output_lines = text.splitlines()
        data_line = output_lines[7]  # headers (nrao and command) is the first 7 lines
        fields = data_line.split()
        lustre_total = fields[1]
        lustre_used = fields[2]

        try:
            # grab the group quota without header information
            capture = subprocess.check_output(['ssh',
                                               self.usage_connector,
                                               'lfs',
                                               'quota',
                                               '-qg',
                                               'vlapipe',
                                               '/lustre/aoc'])
        except subprocess.CalledProcessError:
            # print("Cannot execute qstat command: " + ex.stderr)
            # something went off the rails.
            return None
        # all is well
        text = capture.decode()
        output_lines = text.splitlines()[6]  # skip the NRAO header
        vlapipe_space = output_lines.split()[1]  # used space is the 2nd field

        # populate the output dictionary
        result = defaultdict()
        result["total"] = lustre_total
        result["used"] = lustre_used
        result["vlapipe"] = vlapipe_space

        return result
