""" Test data: ALMA download products """

import json
import warnings
from pathlib import Path

from schema import Project, Author, ExecutionBlock, ScienceProduct
from sqlalchemy import exc as sa_exc

from .expected_values_evla import \
    EXECUTION_BLOCK, METADATA_INGESTION_DATE, CASA_LOG, PPR_FILENAME

EXEC_BLOCK = 'execblock'
CALIBRATED = 'CALIBRATED'
_WEBLOG = 'weblog.tgz'

class AlmaPublicProductSet:
    """ Test data for 2017.1.00297.S """

    def __init__(self):
        self.project = self.build_project()
        self.exec_blocks = self.build_exec_blocks()
        self.science_products = self.build_science_products()
        self.file_info = self.get_file_info()
        self.measurement_set = self.build_basic_ms()
        self.restored_ms = self.build_restored_ms()

    def build_project(self):
        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            project_code = '2017.1.00297.S'
            project = Project(
                project_code=project_code,
                starttime=56799.3877155556,
                endtime=56799.4128683333,
                proprietary_duration=365,
                total_observation_time=0.0583143749972805,
                title='Copy of 2012.1.00060.S for testing',
                abstract = "Understanding the physical factors that control the conversion of interstellar gas into stars is of fundamental importance for both developing a predictive physical theory of star formation and understanding the evolution of galaxies from the earliest epochs of cosmic history to the present time. An important aspect of this question is the study of empirical relations that connect the star formation rate in a given region to local properties of the interstellar medium. An important example is the Schmidt-Kennicutt (KS) law for galaxies that relates the surface densities of the star formation rate and the surface densities of interstellar gas in a non-linear fashion. However, it is also known that there is a linear correlation between the total SFR in galaxies and the mass of dense molecular gas as traced by the high excitation HCN molecule. Contrary to the KS relation, this scaling relation suggests that the total SFR depends simply on the total amount of dense molecular gas in a star forming system. Recently, we have begun to test these scaling relations in the Galactic neighborhood where star formation rates can be much better constrained. We found that for local clouds the total SFR scales most directly, and linearly, with the total mass of high extinction (and dense) molecular gas. Furthermore, we found this linear scaling law between SFR and dense gas to extend and extrapolate directly and smoothly to external galaxies. Moreover, our observations also demonstrate that a KS type relation does not exist for molecular clouds in the Galactic neighborhood. This is a direct consequence of a well known scaling law between the mass and size of molecular clouds, Larson's third law. Overall, our results indicate that a linear scaling law, in which the total amount of dense gas controls the SFR, is the fundamental physical relation that connects star formation across the vast scales from individual GMCs to entire galaxies. Critical testing of these ideas require resolved observations of GMCs in external galaxies. Here we propose to use ALMA to evaluate star formation scaling laws in a nearby galaxy where we can obtain resolved observations of individual GMCs. This allows us to obtain observations of a larger sample of GMCs than is accessible in the Galactic neighborhood. An extensive APEX survey of HII regions in the nearby galaxy NGC 300 has provided us with a sample of 36 star-forming regions with CO(2-1) detections and 42 upper limits. We are currently working on obtaining star formation rates for these regions from multi-wavelength ancillary data including our Herschel observations. We propose to use ALMA's unequalled capabilities to obtain snapshot observations of 40 selected regions in CO(2-1) in order to make resolved measurements of cloud structure to obtain sizes and virial masses. As a pilot project, we also propose to observe the brightest subsample in HCN(1-0) as a dense-gas tracer. Our proposed ALMA CO observations will enable us to to test Larson's scaling laws in an external galaxy and to evaluate which formulation of the Schmidt law is the most meaningful and appropriate to apply to spiral galaxies, and in doing so refine Schmidt's original conjecture of a scaling relation between the rate of star formation and gas density."
            )
            project.authors = [
                Author(project_code=project_code,
                       author_id=32550,
                       username='franzbauer',
                       firstname='Franz',
                       lastname='Bauer',
                       pst_person_id=None,
                       is_pi=True),

                Author(project_code=project_code,
                       author_id=32551,
                       username='ezetre',
                       firstname='Ezequiel',
                       lastname='Treister',
                       pst_person_id=None,
                       is_pi=False),
                Author(project_code=project_code,
                       author_id=32552,
                       username='lho',
                       firstname='Luis',
                       lastname='Ho',
                       pst_person_id=None,
                       is_pi=False),
                Author(project_code=project_code,
                       author_id=32554,
                       username='jyshangguan',
                       firstname='Jinyi',
                       lastname='Shangguan',
                       pst_person_id=None,
                       is_pi=False),
                Author(project_code=project_code,
                       author_id=32553,
                       username='rwang',
                       firstname='Ran',
                       lastname='Wang',
                       pst_person_id=None,
                       is_pi=False),
            ]

            return project

    def build_exec_blocks(self):
        alma_ous_id = 'uid://A001/X1296/Xa93'
        band_code = '05'

        return [
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102209,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=212043,
                ngas_fileset_id='uid___A002_Xcd8029_Xa892',
                starttime=58256.0535855556,
                endtime=58256.1137983333,
                calibration_status=CALIBRATED
            ),
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102169,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=211958,
                ngas_fileset_id='uid___A002_Xcd8029_Xfdd',
                starttime=58255.0338461111,
                endtime=58255.0919594444,
                calibration_status=CALIBRATED
            ),
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102211,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=212047,
                ngas_fileset_id='uid___A002_Xcd8029_Xb0a4',
                starttime=58256.1152683333,
                endtime=58256.1753838889,
                calibration_status=CALIBRATED
            ),
        ]

    def build_science_products(self):
        external_system = 'ALMA Operations'
        return [
            ScienceProduct(
                external_name='uid___A002_Xcd8029_Xfdd',
                filegroup_id=211958,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version='1',
                science_product_type=EXECUTION_BLOCK
            ),
            ScienceProduct(
                external_name='uid___A002_Xcd8029_Xb0a4',
                filegroup_id=212047,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version='1',
                science_product_type=EXECUTION_BLOCK
            ),
            ScienceProduct(
                external_name='uid___A002_Xcd8029_Xa892',
                filegroup_id=212043,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version='1',
                science_product_type=EXECUTION_BLOCK
            ),
        ]

    def get_file_info(self):
        """ Since ALMA files are not in NGAS, we can't use
            location reports to get them; we'll pluck them
            from a JSON file instead.

        """

        # read in contents from JSON
        to_read = None
        for file in Path.cwd().rglob('alma_A001_X1296_Xa93_rawdata.json'):
            to_read = file
            break
        assert to_read is not None
        with open(to_read, 'r') as content:
            file_info = json.loads(content.read())

        return file_info

    def build_basic_ms(self):
        return {
            'uid___A002_Xcd8029_Xa892': {
                'uid___A002_Xcd8029_Xa892.400887655.tar.gz': 504369560,
                'unknown.download_alma_ms.pipeline_manifest.xml': 804,
                'casa_pipescript.py': 680,
                PPR_FILENAME: 4584,
                _WEBLOG: 1592335,
                'unknown.download_alma_ms.pprequest.xml': 4584,
                'unknown.download_alma_ms.auxproducts.tgz': 839,
                CASA_LOG: 1729,
                'pipeline_aquareport.xml': 3472,
                'uid___A002_Xcd8029_Xa892.ms.tgz': 519234359,
            },
            'uid___A002_Xcd8029_Xb0a4': {
                'uid___A002_Xcd8029_Xb0a4.400887655.tar.gz': 503501127,
                'unknown.download_alma_ms.pipeline_manifest.xml': 804,
                CASA_LOG: 1729,
                _WEBLOG: 1576475,
                'casa_pipescript.py': 680,
                'pipeline_aquareport.xml': 3465,
                'uid___A002_Xcd8029_Xb0a4.ms.tgz':  518582184,
                'unknown.download_alma_ms.auxproducts.tgz': 844,
                PPR_FILENAME: 4584,
                'unknown.download_alma_ms.pprequest.xml': 4584,

            },
            'uid___A002_Xcd8029_Xfdd': {
                'uid___A002_Xcd8029_Xfdd.400887655.tar.gz': 489883578,
                'unknown.download_alma_ms.pipeline_manifest.xml': 802,
                'pipeline_aquareport.xml': 3460,
                PPR_FILENAME: 4582,
                'unknown.download_alma_ms.pprequest.xml': 4582,
                _WEBLOG: 1562579,
                'uid___A002_Xcd8029_Xfdd.ms.tgz': 504128830,
                CASA_LOG: 1723,
                'unknown.download_alma_ms.auxproducts.tgz': 836,
                'casa_pipescript.py': 679,
            }
        }

    def build_restored_ms(self):
        return {
            'uid___A002_Xcd8029_Xb0a4': {
                'uid___A002_Xcd8029_Xb0a4.ms.tgz': 1325422495,
                'uid___A002_Xcd8029_Xa892.ms.tgz': 1331336708,
                'uid___A002_Xcd8029_Xfdd.ms.tgz': 1290775660,
                PPR_FILENAME: 5158,

            },
        }


class AlmaImageProduct:
    # TODO -- need to find project w/images
    pass


class AlmaProprietaryProductSet:
    def __init__(self):
        self.project = self.build_project()
        self.exec_blocks = self.build_exec_blocks()
        self.science_products = self.build_science_products()
        self.file_info, self.total_size = self.get_file_info()
        self.measurement_set = self.build_basic_ms()
        self.cms = self.build_cms()

    def build_project(self):
        project = Project(
            # TODO:
        )
        project.authors = self.build_authors()
        return project
